#include <Arduino.h>
#include "Ulm_Logger.h"
#include "Ulm_SDStorage.h"
#include "Ulm_RGB_LED.h"
#include "Ulm_TemperatureSensor_DS18B20.h"
#include "Ulm_OLED_Display.h"
#include "Ulm_LSM6DS3.h"
#include "Ulm_MS5607.h"

// ====================================================================================================================
// SD CARD
// ====================================================================================================================
#define             SD_CS_PIN                                   (10u)
#define             SD_DETECT_PIN                               (2u)

// ====================================================================================================================
// RGB LED (NeoPixel)
// ====================================================================================================================
#define             LED_RGB_PIN                                 (9u)

// ====================================================================================================================
// MS5607 ALTITUDE SENSOR
// ====================================================================================================================
#define             MS5607_ALTITUDE_SENSOR_I2C_ADDRESS          (0x77)

// ====================================================================================================================
// DS18B20 TEMPERATURE SENSOR
// ====================================================================================================================
#define             DS18B20_TEMPERATURE_SENSOR_PIN              (6u)

// ====================================================================================================================
// LSM6DS3 IMU (ACCELEROMETER AND GYROSCOPE)
// ====================================================================================================================
#define             LSM6DS3_IMU_SENSOR_I2C_ADDRESS              (0x6A)

/**
 * TODO
 * Schreibe noch eine Klasse mit Temperatursensor und LED.
 * -> Ulm_SDStorage
 * -> Ulm_StatusLED
 * -> Ulm_TemperatureSensor_DS18B20
 * -> Ulm_Datalogger mit je einem Member der Elemente darüber. Pointerübergabe bzw Argumentübergabe und Pointer nur
 * zu dem SDStorage.
 */

// INPUT DEVICES (SENSORS)
auto imu                = Ulm_LSM6DS3(LSM6DS3_IMU_SENSOR_I2C_ADDRESS);
auto altitudeSensor     = Ulm_MS5607(MS5607_ALTITUDE_SENSOR_I2C_ADDRESS);
auto temperatureSensor  = Ulm_TemperatureSensor_DS18B20(DS18B20_TEMPERATURE_SENSOR_PIN);

// OUTPUT DEVICES (LED AND DISPLAY)
auto led                = Ulm_RGB_LED(LED_RGB_PIN);
auto display            = Ulm_OLED_Display();

// STORAGE DEVICE
Ulm_SDStorage storage = Ulm_SDStorage::getBuilder()
        .atPin(SD_CS_PIN)
        .withDetectPin(SD_DETECT_PIN)
        .atDirectory("DATALOG")
        .withDataFile("DATA_", Ulm_SDStorage::FileType::TXT)
        .withRedundancyFile("RED_", Ulm_SDStorage::FileType::TXT)
        .withLogicForExtremeEnvironments() // SD.open mehrfach -> 2 Sekunden Delay (die lassen sich auch nicht reduzieren)
        .saveSDInformationOnStart()
        .build();


struct SensorData {
    unsigned long   uptimeMillis;
    float           temperature;        // The ambient temperature in Celsius
    float           accelerationX;      // Acceleration for all three axes (g (earth gravity))
    float           accelerationY;
    float           accelerationZ;
    float           gyroscopeX;         // Gyroscope of all three axes (degrees/second)
    float           gyroscopeY;
    float           gyroscopeZ;
    double          pressure;           // Pressure in mBar
    double          altitude;           // Altitude in meters
} sensorData;


void storeSensorData() {
    char buffer[128];
    sprintf(buffer,
            "%lu;%.2f;%.2f;%.2f;%.2f;%.2f;%.2f;%.2f;%.2f;%.2f",
            sensorData.uptimeMillis,
            sensorData.temperature,
            sensorData.accelerationX,
            sensorData.accelerationY,
            sensorData.accelerationZ,
            sensorData.gyroscopeX,
            sensorData.gyroscopeY,
            sensorData.gyroscopeZ,
            sensorData.pressure,
            sensorData.altitude
    );

    if(storage.store(buffer)) {
        led.showSuccess();
    } else {
        led.showError();
    }
}


void setup() {

    uint8_t errors = 0;

    // Start serial interface
    Ulm_Logger::begin();
    Ulm_Logger::printInformation("Starting datalogger 'SGUduino 2023/24...");

    display.begin();
    display.write("Hello!");
    display.display();

    // Start temperature sensor
    temperatureSensor.begin();

    // Start IMU
    if(imu.begin()) {
        Ulm_Logger::printSuccess("Init of IMU done successfully.");
    } else {
        Ulm_Logger::printError("Failed to init IMU!");
        errors++;
    }

    // Start altitude sensor
    if(altitudeSensor.begin()) {
        Ulm_Logger::printSuccess("Init of altitude sensor done successfully.");
    } else {
        Ulm_Logger::printError("Failed to init altitude sensor!");
        errors++;
    }

    // SD Storage
    if(storage.begin()) {
        Ulm_Logger::printSuccess("Init of SD storage done successfully.");
    } else {
        Ulm_Logger::printError("Failed to init SD storage!");
        errors++;
    }

    // Neo-pixel RGB LED
    led.begin();
    led.setBrightness(20);
    led.showSuccess();

    if(errors == 0) {
        Ulm_Logger::printSuccess("Datalogger is ready for take-off! See you soon.");
    } else {
        Ulm_Logger::printWarning(String(errors) + " error(s) occurred. Datalogger might not run correctly!");
    }

    // Write header line of data.
    String header = "";
    header += "Uptime [ms];";
    header += "Temperature [°C];";
    header += "Acceleration X [g];";
    header += "Acceleration Y [g];";
    header += "Acceleration Z [g];";
    header += "Gyroscope X [deg/s];";
    header += "Gyroscope Y [deg/s];";
    header += "Gyroscope Z [deg/s];";
    header += "Pressure [mBar];";
    header += "Altitude [m]";

    if(!storage.store(header)) {
        Serial.println("Failed to write header to SD storage!");
    }
    Serial.println(header);
}


void loop() {

    sensorData.uptimeMillis = millis();

    // Sample temperature sensor
    sensorData.temperature = temperatureSensor.readTemperature();

    // Sample IMU
    imu.readAcceleration(sensorData.accelerationX, sensorData.accelerationY, sensorData.accelerationZ);
    imu.readGyroscope(sensorData.gyroscopeX, sensorData.gyroscopeY, sensorData.gyroscopeZ);

    // Sample altitude sensor
    sensorData.pressure = altitudeSensor.getPressure();
    sensorData.altitude = altitudeSensor.getAltitude();

    // Store data
    storeSensorData();

    // Wait for some time
    delay(1000);
}
